from __future__ import annotations

import functools
import pathlib
import sys

import pytest
import torch

import torchhull

DATA_DIR = pathlib.Path(__file__).parents[1] / "data"
sys.path.append(str(DATA_DIR))
from generate_dataset import generate_dataset  # noqa: E402

generate_dataset = functools.cache(generate_dataset)


DEVICE = torch.device("cuda")


@pytest.mark.parametrize("sparse", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_blurred", [torch.float16, torch.float32, torch.float64])
def test_gaussian_blur_empty_masks(
    sparse: bool,
    dtype_masks: torch.dtype,
    dtype_blurred: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    _, _, masks_reference, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)

    masks_reference = masks_reference.to(dtype=dtype_masks, device=DEVICE)
    masks = torch.zeros_like(masks_reference)

    kernel_size = 5
    sigma = 1

    blurred_masks = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=sparse,
        dtype=dtype_blurred,
    )

    assert blurred_masks.dtype == dtype_blurred
    assert torch.allclose(blurred_masks, masks.to(dtype_blurred))


@pytest.mark.parametrize("sparse", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_blurred", [torch.float16, torch.float32, torch.float64])
def test_gaussian_blur_full_masks(
    sparse: bool,
    dtype_masks: torch.dtype,
    dtype_blurred: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    _, _, masks_reference, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)

    masks_reference = masks_reference.to(dtype=dtype_masks, device=DEVICE)
    masks = torch.ones_like(masks_reference)

    kernel_size = 5
    sigma = 1

    blurred_masks = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=sparse,
        dtype=dtype_blurred,
    )

    assert blurred_masks.dtype == dtype_blurred
    assert torch.allclose(blurred_masks, masks.to(dtype_blurred))


@pytest.mark.parametrize("sparse", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_blurred", [torch.float16, torch.float32, torch.float64])
def test_gaussian_blur_kernel_size_1(
    sparse: bool,
    dtype_masks: torch.dtype,
    dtype_blurred: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    _, _, masks, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)

    masks = masks.to(dtype=dtype_masks, device=DEVICE)

    kernel_size = 1
    sigma = 1

    blurred_masks = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=sparse,
        dtype=dtype_blurred,
    )

    assert blurred_masks.dtype == dtype_blurred
    assert torch.allclose(blurred_masks, masks.to(dtype_blurred))


@pytest.mark.parametrize("kernel_size", [3, 5, 7, 9, 11])
@pytest.mark.parametrize("sigma", [1.0, 100.0])
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("sparse", [True, False])
@pytest.mark.parametrize("dtype_masks", [torch.float16, torch.float32, torch.float64])
def test_gaussian_blur_default_type(
    kernel_size: int,
    sigma: float,
    masks_partial: bool,
    sparse: bool,
    dtype_masks: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    _, _, masks, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)

    masks = masks.to(dtype=dtype_masks, device=DEVICE)

    crop_x = 700
    crop_y = 400
    masks = masks[:, crop_y:-crop_y, crop_x:-crop_x, :].clone() if masks_partial else masks

    blurred_masks = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=sparse,
    )

    assert blurred_masks.dtype == masks.dtype
    assert torch.all(blurred_masks >= 0.0 - 1e-5)
    assert torch.all(blurred_masks <= 1.0 + 1e-5)


@pytest.mark.parametrize("kernel_size", [3, 5, 7, 9, 11])
@pytest.mark.parametrize("sigma", [1.0, 100.0])
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("sparse", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_blurred", [torch.float16, torch.float32, torch.float64])
def test_gaussian_blur(
    kernel_size: int,
    sigma: float,
    masks_partial: bool,
    sparse: bool,
    dtype_masks: torch.dtype,
    dtype_blurred: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    _, _, masks, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)

    masks = masks.to(dtype=dtype_masks, device=DEVICE)

    crop_x = 700
    crop_y = 400
    masks = masks[:, crop_y:-crop_y, crop_x:-crop_x, :].clone() if masks_partial else masks

    blurred_masks = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=sparse,
        dtype=dtype_blurred,
    )

    assert blurred_masks.dtype == dtype_blurred
    assert torch.all(blurred_masks >= 0.0 - 1e-5)
    assert torch.all(blurred_masks <= 1.0 + 1e-5)


@pytest.mark.parametrize("kernel_size", [3, 5, 7, 9, 11])
@pytest.mark.parametrize("sigma", [1.0, 100.0])
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_blurred", [torch.float16, torch.float32, torch.float64])
def test_gaussian_blur_consistent(
    kernel_size: int,
    sigma: float,
    masks_partial: bool,
    dtype_masks: torch.dtype,
    dtype_blurred: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    _, _, masks, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)

    masks = masks.to(dtype=dtype_masks, device=DEVICE)

    crop_x = 700
    crop_y = 400
    masks = masks[:, crop_y:-crop_y, crop_x:-crop_x, :].clone() if masks_partial else masks

    blurred_masks_dense = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=False,
        dtype=dtype_blurred,
    )

    blurred_masks_sparse = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=True,
        dtype=dtype_blurred,
    )

    if dtype_blurred == torch.float16:
        epsilon = 3e-3
    if dtype_blurred == torch.float32:
        epsilon = 1e-7
    if dtype_blurred == torch.float64:
        epsilon = 1e-7

    assert torch.allclose(blurred_masks_dense, blurred_masks_sparse, atol=epsilon)



from __future__ import annotations

import enum

import pytest
import torch

import torchhull

DEVICE = torch.device("cuda")


def sdf_sphere(center: torch.Tensor, radius: float, samples: torch.Tensor) -> torch.Tensor:
    return torch.linalg.norm(samples - torch.unsqueeze(center, 0), dim=1) - radius


def is_closed_manifold(faces: torch.Tensor) -> bool:
    assert faces.shape[1] == 3  # noqa: PLR2004

    edges = torch.cat([faces[:, (0, 1)], faces[:, (1, 2)], faces[:, (2, 0)]], dim=0)

    edges_ordered, _ = torch.sort(edges, dim=1)
    _, edges_counts = torch.unique(edges_ordered, dim=0, return_counts=True)

    return torch.equal(edges_counts, torch.full_like(edges_counts, 2))


def list_sizes() -> list[int]:
    return [100, 200, 300, 400]


class TensorType(enum.Enum):
    DENSE = enum.auto()
    SPARSE = enum.auto()

    def __str__(self) -> str:
        return f"{self.name}"


@pytest.mark.parametrize("size", list_sizes())
@pytest.mark.parametrize("tensor_type", [TensorType.DENSE, TensorType.SPARSE])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_marching_cubes_empty_input(
    size: int,
    tensor_type: TensorType,
    dtype: torch.dtype,
) -> None:
    if tensor_type == TensorType.DENSE:
        sdf = torch.full([1, *(3 * [size])], -1.0, dtype=dtype, device=DEVICE)
    elif tensor_type == TensorType.SPARSE:
        sdf = torch.sparse_coo_tensor(
            indices=torch.empty([4, 0], dtype=torch.int64, device=DEVICE),
            values=torch.empty([0], dtype=dtype, device=DEVICE),
            size=[1, *(3 * [size])],
        )

    verts, faces = torchhull.marching_cubes(sdf, isolevel=0, return_local_coords=False)

    assert verts.size(0) == 0
    assert faces.size(0) == 0


@pytest.mark.parametrize("size", list_sizes())
@pytest.mark.parametrize("tensor_type", [TensorType.DENSE, TensorType.SPARSE])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_marching_cubes(
    size: int,
    tensor_type: TensorType,
    dtype: torch.dtype,
) -> None:
    grid_1d = torch.arange(size, device=DEVICE)

    grid_verts = torch.stack(
        torch.meshgrid(3 * [grid_1d], indexing="ij"),
        dim=0,
    )  # 3 x size x size x size
    grid_verts = grid_verts.reshape([3, -1]).T.to(dtype=dtype)  # size^3 x 3

    center = torch.full((3,), size / 2, dtype=dtype, device=DEVICE)
    radius = 0.8731945 * (size / 2)  # Use very uneven fraction to avoid zero values at the vertices
    sdf = sdf_sphere(center, radius, grid_verts)

    sdf = sdf.reshape([1, *(3 * [size])])

    if tensor_type == TensorType.SPARSE:
        truncation_distance = 2  # max(3, 0.01 * max(sdf.shape))
        sdf[sdf.abs() > truncation_distance] = 0

        sdf = sdf.to_sparse()

    verts, faces = torchhull.marching_cubes(sdf, isolevel=0, return_local_coords=False)

    sdf_verts = sdf_sphere(center, radius, verts)
    assert torch.allclose(sdf_verts, torch.zeros_like(sdf_verts), atol=1e-2)
    assert is_closed_manifold(faces)



from __future__ import annotations

import functools
import pathlib
import sys

import pytest
import torch

import torchhull

DATA_DIR = pathlib.Path(__file__).parents[1] / "data"
sys.path.append(str(DATA_DIR))
from generate_dataset import generate_dataset  # noqa: E402

generate_dataset = functools.cache(generate_dataset)


DEVICE = torch.device("cuda")


def is_closed_manifold(faces: torch.Tensor) -> bool:
    assert faces.shape[1] == 3  # noqa: PLR2004
    edges = torch.cat([faces[:, (0, 1)], faces[:, (1, 2)], faces[:, (2, 0)]], dim=0)

    edges_ordered, _ = torch.sort(edges, dim=1)
    _, edges_counts = torch.unique(edges_ordered, dim=0, return_counts=True)

    return torch.equal(edges_counts, torch.full_like(edges_counts, 2))


def list_levels() -> list[int]:
    return list(range(1, 8 + 1))


@pytest.mark.parametrize("level", list_levels())
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("unique_verts", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_transforms", [torch.float32, torch.float64])
def test_visual_hull_empty_masks(
    level: int,
    masks_partial: bool,
    unique_verts: bool,
    dtype_masks: torch.dtype,
    dtype_transforms: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    projection_matrices, view_matrices, masks_reference, _, _, _ = generate_dataset(
        mesh_file=data_dir / file, device=DEVICE
    )
    transforms = projection_matrices @ view_matrices

    masks_reference = masks_reference.to(dtype=dtype_masks, device=DEVICE)
    masks = torch.zeros_like(masks_reference)

    transforms = transforms.to(dtype=dtype_transforms, device=DEVICE)

    scale = 1.1

    vertices, faces = torchhull.visual_hull(
        masks=masks,
        transforms=transforms,
        level=level,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
        masks_partial=masks_partial,
        transforms_convention="opengl",
        unique_verts=unique_verts,
    )

    assert vertices.size(0) == 0
    assert faces.size(0) == 0


@pytest.mark.parametrize("level", list_levels())
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("unique_verts", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_transforms", [torch.float32, torch.float64])
def test_visual_hull_full_masks(
    level: int,
    masks_partial: bool,
    unique_verts: bool,
    dtype_masks: torch.dtype,
    dtype_transforms: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    projection_matrices, view_matrices, masks_reference, _, _, _ = generate_dataset(
        mesh_file=data_dir / file, device=DEVICE
    )
    transforms = projection_matrices @ view_matrices

    masks_reference = masks_reference.to(dtype=dtype_masks, device=DEVICE)
    masks = torch.ones_like(masks_reference)

    transforms = transforms.to(dtype=dtype_transforms, device=DEVICE)

    scale = 1.1

    vertices, faces = torchhull.visual_hull(
        masks=masks,
        transforms=transforms,
        level=level,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
        masks_partial=masks_partial,
        transforms_convention="opengl",
        unique_verts=unique_verts,
    )

    assert vertices.size(0) == 0
    assert faces.size(0) == 0


@pytest.mark.parametrize("level", list_levels())
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("unique_verts", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_transforms", [torch.float32, torch.float64])
def test_visual_hull(
    level: int,
    masks_partial: bool,
    unique_verts: bool,
    dtype_masks: torch.dtype,
    dtype_transforms: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    projection_matrices, view_matrices, masks, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)
    transforms = projection_matrices @ view_matrices

    masks = masks.to(dtype=dtype_masks, device=DEVICE)
    transforms = transforms.to(dtype=dtype_transforms, device=DEVICE)

    scale = 1.1

    vertices, faces = torchhull.visual_hull(
        masks=masks,
        transforms=transforms,
        level=level,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
        masks_partial=masks_partial,
        transforms_convention="opengl",
        unique_verts=unique_verts,
    )

    assert torch.all(vertices >= torch.tensor([[-scale, -scale, -scale]], dtype=vertices.dtype, device=vertices.device))
    assert torch.all(vertices <= torch.tensor([[scale, scale, scale]], dtype=vertices.dtype, device=vertices.device))
    if unique_verts:
        assert is_closed_manifold(faces)
    else:
        assert vertices.size(0) == 3 * faces.size(0)


@pytest.mark.parametrize("level", list(range(1, 5 + 1)))  # NN search is slow, so only test low levels
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("unique_verts", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_transforms", [torch.float32, torch.float64])
def test_visual_hull_transforms_conventions(
    level: int,
    masks_partial: bool,
    unique_verts: bool,
    dtype_masks: torch.dtype,
    dtype_transforms: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    (
        projection_matrices,
        view_matrices,
        masks,
        projection_matrices_cv,
        view_matrices_cv,
        masks_cv,
    ) = generate_dataset(mesh_file=data_dir / file, device=DEVICE)
    transforms = projection_matrices @ view_matrices
    transforms_cv = projection_matrices_cv @ view_matrices_cv

    masks = masks.to(dtype=dtype_masks, device=DEVICE)
    transforms = transforms.to(dtype=dtype_transforms, device=DEVICE)

    scale = 1.1

    vertices, faces = torchhull.visual_hull(
        masks=masks,
        transforms=transforms,
        level=level,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
        masks_partial=masks_partial,
        transforms_convention="opengl",
        unique_verts=unique_verts,
    )

    assert torch.all(vertices >= torch.tensor([[-scale, -scale, -scale]], dtype=vertices.dtype, device=vertices.device))
    assert torch.all(vertices <= torch.tensor([[scale, scale, scale]], dtype=vertices.dtype, device=vertices.device))
    if unique_verts:
        assert is_closed_manifold(faces)
    else:
        assert vertices.size(0) == 3 * faces.size(0)

    vertices_cv, faces_cv = torchhull.visual_hull(
        masks=masks_cv,
        transforms=transforms_cv,
        level=level,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
        masks_partial=masks_partial,
        transforms_convention="opencv",
        unique_verts=unique_verts,
    )
    vertices_cv = (
        torch.tensor([1, -1, -1], dtype=transforms_cv.dtype, device=transforms_cv.device).diag() @ vertices_cv.T
    ).T  # CV -> GL

    assert torch.all(
        vertices_cv >= torch.tensor([[-scale, -scale, -scale]], dtype=vertices_cv.dtype, device=vertices_cv.device)
    )
    assert torch.all(
        vertices_cv <= torch.tensor([[scale, scale, scale]], dtype=vertices_cv.dtype, device=vertices_cv.device)
    )
    if unique_verts:
        assert is_closed_manifold(faces_cv)
    else:
        assert vertices_cv.size(0) == 3 * faces_cv.size(0)

    # This is slow but memory-efficient
    epsilon = 1e-4
    for i in range(vertices.shape[0]):
        nn_distances = torch.linalg.norm(vertices_cv - vertices[i : (i + 1)], dim=1)
        nn = torch.topk(nn_distances, k=1, largest=False)
        assert nn.values[0] < epsilon


@pytest.mark.parametrize("level", list_levels())
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("unique_verts", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float16, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_transforms", [torch.float32, torch.float64])
def test_visual_hull_with_blur(
    level: int,
    masks_partial: bool,
    unique_verts: bool,
    dtype_masks: torch.dtype,
    dtype_transforms: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    projection_matrices, view_matrices, masks, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)
    transforms = projection_matrices @ view_matrices

    masks = masks.to(dtype=dtype_masks, device=DEVICE)
    transforms = transforms.to(dtype=dtype_transforms, device=DEVICE)

    kernel_size = 9
    sigma = 2
    masks = torchhull.gaussian_blur(
        images=masks,
        kernel_size=kernel_size,
        sigma=sigma,
        sparse=True,
        dtype=dtype_transforms,
    )

    scale = 1.1

    vertices, faces = torchhull.visual_hull(
        masks=masks,
        transforms=transforms,
        level=level,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
        masks_partial=masks_partial,
        transforms_convention="opengl",
        unique_verts=unique_verts,
    )

    assert torch.all(vertices >= torch.tensor([[-scale, -scale, -scale]], dtype=vertices.dtype, device=vertices.device))
    assert torch.all(vertices <= torch.tensor([[scale, scale, scale]], dtype=vertices.dtype, device=vertices.device))
    if unique_verts:
        assert is_closed_manifold(faces)
    else:
        assert vertices.size(0) == 3 * faces.size(0)


@pytest.mark.parametrize("level", list_levels())
@pytest.mark.parametrize("masks_partial", [True, False])
@pytest.mark.parametrize("unique_verts", [True, False])
@pytest.mark.parametrize(
    "dtype_masks", [torch.int8, torch.int16, torch.int32, torch.int64, torch.float32, torch.float64]
)
@pytest.mark.parametrize("dtype_transforms", [torch.float32, torch.float64])
def test_visual_hull_with_candidate_voxels(
    level: int,
    masks_partial: bool,
    unique_verts: bool,
    dtype_masks: torch.dtype,
    dtype_transforms: torch.dtype,
) -> None:
    data_dir = pathlib.Path(__file__).parents[1] / "data"
    file = "Armadillo.ply"

    projection_matrices, view_matrices, masks, _, _, _ = generate_dataset(mesh_file=data_dir / file, device=DEVICE)
    transforms = projection_matrices @ view_matrices

    masks = masks.to(dtype=dtype_masks, device=DEVICE)
    transforms = transforms.to(dtype=dtype_transforms, device=DEVICE)

    scale = 1.1

    vertices, faces, candidates = torchhull.visual_hull_with_candidate_voxels(
        masks=masks,
        transforms=transforms,
        level=level,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
        masks_partial=masks_partial,
        transforms_convention="opengl",
        unique_verts=unique_verts,
    )

    wireframes = torchhull.candidate_voxels_to_wireframes(
        candidates,
        cube_corner_bfl=(-scale, -scale, -scale),
        cube_length=2.0 * scale,
    )

    assert torch.all(vertices >= torch.tensor([[-scale, -scale, -scale]], dtype=vertices.dtype, device=vertices.device))
    assert torch.all(vertices <= torch.tensor([[scale, scale, scale]], dtype=vertices.dtype, device=vertices.device))
    if unique_verts:
        assert is_closed_manifold(faces)
    else:
        assert vertices.size(0) == 3 * faces.size(0)

    for i, (w_vertices, _) in enumerate(wireframes):
        assert torch.all(
            w_vertices >= torch.tensor([-scale, -scale, -scale], dtype=w_vertices.dtype, device=w_vertices.device)
        )
        assert torch.all(
            w_vertices <= torch.tensor([scale, scale, scale], dtype=w_vertices.dtype, device=w_vertices.device)
        )

        if i > 0:
            w_previous_vertices, _ = wireframes[i - 1]
            bb_min, _ = torch.min(w_previous_vertices, dim=0)
            bb_max, _ = torch.max(w_previous_vertices, dim=0)
            assert torch.all(w_vertices >= bb_min)
            assert torch.all(w_vertices <= bb_max)



"""
torchhull is an extremely fast Torch C++/CUDA implementation for computing visual hulls from mask images and comes
with Python bindings through `charonload <https://github.com/vc-bonn/charonload>`_:

.. code-block:: python

    import torchhull

    verts, faces = torchhull.visual_hull(masks,  # [B, H, W, 1]
                                         transforms,  # [B, 4, 4]
                                         level,
                                         cube_corner_bfl,
                                         cube_length,
                                         masks_partial=False,
                                         unique_verts=True,
                                        )
"""

import sys

required_version = (3, 9)

if sys.version_info[:2] < required_version:  # pragma: no cover
    msg = "%s requires Python %d.%d+" % (__package__, *required_version)  # noqa: UP031
    raise RuntimeError(msg)

del required_version
del sys


import email.utils
import importlib.metadata
import pathlib

import charonload

PROJECT_ROOT_DIRECTORY = pathlib.Path(__file__).parents[2]

VSCODE_STUBS_DIRECTORY = PROJECT_ROOT_DIRECTORY / "typings"


charonload.module_config["_c_torchhull"] = charonload.Config(
    pathlib.Path(__file__).parent / "_C",
    stubs_directory=VSCODE_STUBS_DIRECTORY,
    verbose=False,
)


from _c_torchhull import (
    candidate_voxels_to_wireframes,
    gaussian_blur,
    marching_cubes,
    sparse_visual_hull_field,
    store_curve_network,
    visual_hull,
    visual_hull_with_candidate_voxels,
)

__author__ = ", ".join(
    [
        email.utils.parseaddr(author.strip())[0]
        for author in importlib.metadata.metadata(__package__)["Author-email"].split(",")
    ]
)
__version__ = importlib.metadata.version(__package__)
__copyright__ = f"2024, {__author__}"

__all__ = [
    "candidate_voxels_to_wireframes",
    "gaussian_blur",
    "marching_cubes",
    "sparse_visual_hull_field",
    "store_curve_network",
    "visual_hull",
    "visual_hull_with_candidate_voxels",
]
